/* ccb.c: a C beautifier (re-indenter/formatter) program
 *  *  Author : Dr. Charles E. Campbell, Jr.
 *   *
 *    *  History:
 *     *    v1.07 :  1/27/2000 - includes private/public/protected detab exception
 *      *            10/ 5/2000 - made nearly all functions static
 *       *    v1.06 :  6/12/1997 - now allows leading whitespace in from
 *        *                         of preprocessor symbol "#"
 *         *
 *          */
#include <stdio.h>
#include <ctype.h>
#include "xtdio.h"

/* ------------------------------------------------------------------------
 *  * Definitions:
 *   */
#define BUFSIZE		256
#define BUFOVER		10
#define OUTBUFSIZE	4096
#define NXTNRML		0
#define NXTBEGIN	1

#define CCBNRML			0
#define CCBSTRING		1
#define CCBCOMMENT		2
#define CCBPREPRO		4
#define CCBBAR			8
#define CCBIFWHILE		16
#define CCBDO			32

#define NOTFUNC			0
#define MAYBEFUNC		1
#define INFUNC			2

#define BFLAG_DFLT		72
#define CSKIP			((char) 1)

/* ------------------------------------------------------------------------
 *  * Typedefs:
 *   */
/* typedef struct environ_str   Environ;
typedef struct lookahead_str LookAhead;
*/
/* ------------------------------------------------------------------------
 *  * Local Data Structures:
 *   */
struct environ_str {
		char     string;
			char    *tab;
				char     tabs[BUFSIZE];
					int      cmmntbar;
						int      curly;
							int      func;			/* NOTFUNC  MAYBEFUNC  INFUNC	*/
								int      paren;
									unsigned wierd;
										Environ *nxt,*prv;
											};

struct lookahead_str {
		char      *buf;
			LookAhead *nxt,*prv;
				};

/* ------------------------------------------------------------------------
 *  * Options:
 *   */
static int bflag   = 0;				/* place a comment bar between functions			*/
static int Bflag   = 0;				/* replace leading tabs and blanks, not just tabs	*/
static int cflag   = 1;				/* detab case statements by one						*/
static int dflag   = 0;				/* detab close curlys by one tab					*/
static int fflag   = 1;				/* detab inside of functions by one tab				*/
static int iflag   = 1;				/* if/do/while processing							*/
static int sflag   = 0;				/* save output in a *.cb file						*/
static int rflag   = 0;				/* replace input file with beautified file			*/
static int vflag   = 0;				/* verbose mode										*/
static int wflag   = 0;				/* warning only mode								*/

/* ------------------------------------------------------------------------
 * * Global Data:
 * */
char      *b       = NULL;			/* kept global for access to b[1], ...				*/
char       buf[BUFSIZE + BUFOVER];	/* input buffer										*/
char       infname[BUFSIZE];		/* input filename									*/
char       outbuf[OUTBUFSIZE];		/* output buffer									*/
char      *outb    = outbuf;		/* output buffer									*/
char       outfname[BUFSIZE];		/* output filename									*/
int        cppmode = 0;				/* c++ mode (input file is C++)						*/
long       linenum = 0L;			/* maintained by nxtc, supports error messages		*/
FILE      *fp      = NULL;			/* kept global, sigh, for speed						*/
FILE      *fpout   = NULL;			/* output FILE handle								*/
LookAhead *lhd     = NULL;			/* supports lookahead/nxtc							*/
LookAhead *ltl     = NULL;			/* supports lookahead/nxtc							*/

/* ------------------------------------------------------------------------
 *  * Explanation:
 *   */
char *usage[]= {
		XYELLOW,"ccb",XGREEN," [",XCYAN,"-bBcdfirsv",XGREEN,"]\n\n",

			XWHITE,"Copyright 1993,1997,2000 by Charles E. Campbell, Jr., Ph.D.\n",
				"All rights reserved except for those explicitly granted in the\n",
					"file <",XRVGREEN,"Copyright",XWHITE,"> that accompanied this program\n",
						XUWHITE,"Use at your own risk!\n\n",

							"\n",XWHITE,"Options\n",
								XCYAN,"  -b ",XGREEN,": place a bar between functions\n",
									XCYAN,"  -B ",XGREEN,": replace leading blanks+tabs, not just tabs\n",
										XCYAN,"  -BB",XGREEN,": replace leading tabs and groups of blanks >= 3\n",
											XCYAN,"  -c ",XGREEN,": don't detab case statements by one\n",
												XCYAN,"  -d ",XGREEN,": de-tab close curly\n",
													XCYAN,"  -f ",XGREEN,": don't de-tab functions by one\n",
														XCYAN,"  -i ",XGREEN,": don't use if/do/while/else/for mode\n",
															XCYAN,"  -r ",XGREEN,": replace input file with beautified file (dangerous)\n",
																XCYAN,"  -s ",XGREEN,": save output in *.cb file(s)\n",
																	XCYAN,"  -v ",XGREEN,": verbose mode (to stderr)\n",
																		XCYAN,"  -w ",XGREEN,": only warnings mode\n",

																			"\n",XWHITE,"Explanation",XGREEN,"\n",

																				"  The ",XYELLOW,"ccb",XGREEN," program is Dr. Chip's C beautifier program.\n",
																					"  It enforces tabbing rules based on the current number of open\n",
																						"  curly braces and can shift comments and open curly braces around\n",
																							"  near if, for, else, else if, do, and switch statements.  The ccb\n",
																								"  program has an obvious relationship with the U*IX cb program, but\n",
																									"  has a different (default) style.  The default style is, of course,\n",
																										"  Dr. Chip's preference, but can be modified to the \"standard\" cb\n",
																											"  style with \"-cdfi\".\n\n",

																												"  Dr. Chip's C beautifier also contains extensions which should enable\n",
																													"  it to work with C++ programs, too, although I haven't tested it\n",
																														"  extensively for that purpose (yet).\n",

																															"\n",XWHITE,"Author",XGREEN,"   : Dr. Charles E. Campbell, Jr.\n",
																																     XWHITE,"Date",XGREEN,"     : January 27, 2000\n",
																																     	     XWHITE,"Version",XGREEN,"  : 1.07\n",
																																	     	     XWHITE,"Copyright",XGREEN,": 1999, 2000  Charles E. Campbell, Jr.\n",
																																		     	"",""};

/* ------------------------------------------------------------------------
 *  * Prototypes:
 *   */
#ifdef __PROTOTYPE__
int main( int, char **);             /* ccb.c */
static int options(char **,char **); /* ccb.c */
static void fprior(int,PARMDEF *);   /* ccb.c */
static void dofile(char *);          /* ccb.c */
static void setupfiles(char *);      /* ccb.c */
static int nxtc(void);               /* ccb.c */
static char *smartskip(char *);      /* ccb.c */
static int fixcurly(Environ *);      /* ccb.c */
static void ccb(void);               /* ccb.c */
static void indent(char,Environ *);  /* ccb.c */
static void pushenv(Environ *);      /* ccb.c */
static void popenv(void);            /* ccb.c */
static Environ accessenv(void);      /* ccb.c */
static void output(void);            /* ccb.c */
static void initenv(Environ *);      /* ccb.c */
#else
extern int main();                   /* ccb.c */
static int options();                /* ccb.c */
static void fprior();                /* ccb.c */
static void dofile();                /* ccb.c */
static void setupfiles();            /* ccb.c */
static int nxtc();                   /* ccb.c */
static char *smartskip();            /* ccb.c */
static int fixcurly();               /* ccb.c */
static void ccb();                   /* ccb.c */
static void indent();                /* ccb.c */
static void pushenv();               /* ccb.c */
static void popenv();                /* ccb.c */
static Environ accessenv();          /* ccb.c */
static void output();                /* ccb.c */
static void initenv();               /* ccb.c */
#endif

/* ------------------------------------------------------------------------
 *  * Source Code:
 *   */

/* main: */
int main(
		  int argc,
		    char **argv)
{
	/* initialize Dr. Campbell's debugging system */

	/* initialize color escape sequences */
	rdcolor();

	/* call upon the command line parser */
	getarg(argc,argv,0,(PARMDEF *) NULL,usage,options,fprior,dofile,1,"Enter filename(s): ");

#ifdef vms
	return 1;
#else
	return 0;
#endif
}

/* ------------------------------------------------------------------------ */

/* options: this file processes options */
static int options(char **s,char **a)
{
	int ret=0;


	switch(**s) {

		case 'b':	/* place a comment bar between functions */
				if(isdigit(s[0][1])) {
							sscanf(s[0]+1,"%d",&bflag);
									for(; s[0][ret]; ++ret) if(!isdigit(s[0][ret])) break;
											ret= -ret;	/* tell getarg to skip "|ret|" characters */
													}
					else if(!bflag) bflag= BFLAG_DFLT;
						else            bflag= 0;
							break;

		case 'B':	/* replace leading blanks+tabs, not just tabs */
								if(++Bflag >= 3) Bflag= 0;
									break;

		case 'c':	/* detab case statements by one */
										cflag= !cflag;
											break;

		case 'd':	/* detab close curly */
												dflag= !dflag;
													break;

		case 'f':	/* 1: de-tab functions by one, 0: else */
														fflag= !fflag;
															break;

		case 'i':	/* if/do/while processing */
																iflag= !iflag;
																	break;

		case 'r':	/* replace input file with output file flag (DANGEROUS!) */
																		rflag= !rflag;
																			if(rflag) sflag= 1;
																				break;

		case 's':	/* save output in a *.cb file (if true) */
																					if(rflag) sflag= 1;
																						else      sflag= !sflag;
																							break;

		case 'v':	/* verbose flag */
																								vflag= !vflag;
																									break;

		case 'w':	/* warning only mode */
																										wflag= 1;
																											break;

		default:
																												error(XTDIO_WARNING,"option <%s%s%s> not currently supported\n",RVGREEN,cprt(**s),GREEN);
																													break;
																														}

	return ret;
}

/* ------------------------------------------------------------------------ */

/* fprior: this function processes prior to dofile */
static void fprior(int parmqty,PARMDEF *parms)
{

	fprintf(stderr,"%sCopyright 1993 by Charles E. Campbell, Jr., Ph.D.\n",WHITE);
	fprintf(stderr,"Use at your own risk.%s\n",NRML);

}

/* ------------------------------------------------------------------------ */

/* dofile: this function processes the command line filenames */
static void dofile(char *filename)
{

	/* set up file handles and filenames */
	setupfiles(filename);

	/* beautify input file (only if setupfiles succeeded!) */
	if(fp && fpout) {

			ccb();		/* beautify the C source file		*/
				b= NULL;	/* reset b to null for next file	*/

					/* Close the file handles */
					if(fp != stdin)     fclose(fp);
						if(fpout != stdout) fclose(fpout);

							/* Replace input file with output file (dangerous!) */
							if(rflag) {
										if(mkbkupfile(infname,outfname)) {
														error(XTDIO_ERROR,"unable to make a backup file from <%s%s%s>\n",YELLOW,sprt(infname),GREEN);
																	}
												}
								}

}

/* ------------------------------------------------------------------------ */

/* setupfiles: this function sets up the input and output file handles
 *  * and filenames
 *   */
static void setupfiles(char *filename)
{

	/* handle "stdin" as standard input */
	if(!strcmp(filename,"stdin")) {
			fp   = stdin;
				fpout= stdout;
					strcpy(infname,"stdin");
						strcpy(outfname,"stdout");
							rflag= 0;
								}
	else {
			char *dot;
				char *basename;

					basename= strrchr(filename,'/'); /* point past leading path     */
					    if(basename) ++basename;
					        else         basename= filename;
							dot= strchr(basename,'.');       /* point to first dot (suffix) */

								/* figure out what the input filename is */
								fp= dot? fopen(filename,"r") : NULL;

									if(fp) {	/* opened file as is */
												strcpy(infname,filename);
														if(!sflag) {
																		strcpy(outfname,"stdout");
																					fpout= stdout;
																								}
																else {
#if defined(MSDOS) || defined(vms)
																				dot= strchr(infname,'.');
																							if(dot) *dot= '\0';
																										sprintf(outfname,"%s.cb",infname);
																													*dot= '.';
#else
																																sprintf(outfname,"%s.cb",infname);
#endif
																																			fpout= fopen(outfname,"w");
																																						if(!fpout)
																																										  error(XTDIO_ERROR,"unable to open <%s> for output\n",sprt(outfname));
																																									}
																		}

										else {	/* attempt to infer a suffix */
													int           isfx;
															static char *sfxtbl[]={"c","C","cpp","c++","cp","cxx",""};

																	for(isfx= 0; sfxtbl[isfx][0]; ++isfx) {
																					sprintf(infname,"%s.%s",basename,sfxtbl[isfx]);
																								fp= fopen(infname,"r");
																											if(fp) {
																																if(!sflag) {
																																						strcpy(outfname,"stdout");
																																											fpout= stdout;
																																																}
																																				else {
#if defined(MSDOS) || defined(vms)
																																										dot= strchr(infname,'.');
																																															if(dot) *dot= '\0';
																																																				sprintf(outfname,"%s.cb",infname);
																																																									*dot= '.';
#else
																																																														sprintf(outfname,"%s.cb",infname);
#endif
																																																																			fpout= fopen(outfname,"w");
																																																																								if(!fpout)
																																																																														  error(XTDIO_ERROR,"unable to open <%s> for output\n",
																																																																																  					  sprt(outfname));
																																																																													}
																																								break;
																																												}
																														}
																			/* determine if the input file is a C++ file */
																			cppmode= isfx > 0;
																					}
											}

}

/* ------------------------------------------------------------------------ */

/* nxtc: returns next character (ignore leading tabs/spaces according to Bflag) */
static int nxtc()
{

	if(!b) {	/* first line read in file */
			if(!fgets(buf,BUFSIZE,fp)) {
						b= NULL;
								return EOF;
										}

				/* skip leading tabs (and, optionally, blanks) */
				switch(Bflag) {

						case 1:	/* elide all leading blanks or tabs */
									for(b= buf; *b == '\t' || *b == ' '; ++b) ;
											break;

												case 2:	/* elide all leading tabs, and any qty blanks > 3 */
													for(b= buf; *b == '\t' || *b == ' '; ++b) {
																	if(b[0] == ' ') {
																						if((b[1] != ' ' || b[2] != ' ')) break;
																										for(; *b == ' '; ++b) ;
																														--b;
																																		}
																				}
															break;

																default:
																	b= smartskip(buf);
																			break;
																					}
					if(!*b && b > buf) --b;

						linenum= 1L;
							}

	else if(!*++b) {						/* end of a line in the file	*/

			if(lhd) { 							/* use LookAhead buffer			*/
						if(lhd) {
										char *bb,*lbb;
													for(bb= buf, lbb= lhd->buf; *lbb; ++lbb) if(*lbb != CSKIP) *(bb++)= *lbb;
																*bb= '\0';
																			if(lhd->buf) free(lhd->buf);
																						lhd->buf= NULL;
																									delete_double_link(LookAhead,lhd,lhd,ltl);
																												}
								else if(!fgets(buf,BUFSIZE,fp)) {
												b= NULL;
															return EOF;
																		}
										}

				else if(!fgets(buf,BUFSIZE,fp)) {	/* get line from file			*/
							b= NULL;
									return EOF;
											}

					/* skip leading tabs (and, optionally, blanks) */
					switch(Bflag) {

							case 1:	/* elide all leading blanks or tabs */
										for(b= buf; *b == '\t' || *b == ' ' || *b == CSKIP; ++b) ;
												break;

													case 2:	/* elide all leading tabs, and any qty blanks > 3 */
														for(b= buf; *b == '\t' || *b == ' ' || *b == CSKIP; ++b) {
																		if(b[0] == ' ') {
																							if((b[1] != ' ' || b[2] != ' ')) break;
																											for(; *b == ' '; ++b) ;
																															--b;
																																			}
																					}
																break;

																	default:
																		b= smartskip(buf);
																				break;
																						}
						if(!*b) --b;

							++linenum;
								}

	else {
			while(*b == CSKIP) ++b;
				return NXTNRML;
					}

	return NXTBEGIN;
}

/* ------------------------------------------------------------------------ */

/* smartskip: this function skips past intermixed tabs and blanks to just
 *  * past the last tab
 *   */
static char *smartskip(char *buf)
{
	char *b;


	for(b= buf; *b == '\t' || *b == ' ' || *b == CSKIP; ++b) if(*b == '\t') buf= b + 1;

	return buf;
}

/* --------------------------------------------------------------------- */

/* fixcurly: supports keeping open curlies on the same line as if/while/do
 *  *  by looking ahead to the next character which is
 *   *    1) not a comment, and
 *    *    2) not white space
 *     */
static int fixcurly(Environ *environ)
{
	char    *look             = b;
	char     lookbuf[BUFSIZE];
	int      fixed            = 0;
	Environ  env;


	if(!look) {
			return fixed;
				}

	initenv(&env);

	while(1) {

			/* get look pointed to next character */
			if(*look && *look != '\n') ++look;
				else if(!fgets(lookbuf,BUFSIZE,fp)) {
							return fixed;
									}
					else {	/* generate a new LookAhead structure */

								double_link(LookAhead,lhd,ltl,"fixcurly");

										ltl->buf= (char *)calloc((size_t) strlen(lookbuf)+BUFOVER,sizeof(char));
												outofmem(ltl->buf,"allocating lookahead buffer with <%s>\n",sprt(lookbuf));

														strcpy(ltl->buf,lookbuf);
																look= ltl->buf;
																		}

						if(isspace(*look)) continue;
							if(env.wierd & CCBCOMMENT) {
										if(look[0] == '*' && look[1] == '/') {
														++look;
																	env.wierd&= ~CCBCOMMENT;
																				}
												}
								else if(look[0] == '/') {
											if(look[1] == '*') {
															++look;
																		env.wierd|= CCBCOMMENT;
																					}
													else if(look[1] == '/') {	/* C++ style comment, skip to end-of-line */
																	look+= strlen(look);
																				}
															}
									else break;
										}

	/* put the close curly in the "proper" place */
	if(*look == '{') /* } */ {
			char      *future;
				LookAhead *lookahead=ltl;

					(*outb++)= ' ';
						(*outb++)= '{';	/* } */
							fixed    = 1;

								*look = CSKIP;
									future= stpblk(look+1);
										if(!*future) {
													look[1]= '\n';
															look[2]= '\0';
																	}
	else if(*future != '/') {
				(*outb++)= '\n';
						strcpy(outb,environ->tabs);
								outb    += (environ->tab - environ->tabs);
										(*outb++)= '\t';
												}

	/* look back */
	lookahead= ltl;
		while(1) {
					if(look == buf || (lookahead && look == lookahead->buf)) {
									lookahead= lookahead->prv;
												if(!lookahead) break;
															look     = lookahead->buf + strlen(lookahead->buf) - 1;
																		}
							else --look;
									if(isspace(*look)) *look= CSKIP;
											else               break;
													};

	}

return fixed;
}

/* ------------------------------------------------------------------------ */

/* ccb: Dr. Chip's C and C++ beautifier */
static void ccb()
{
	char    *nb;          /* first not-blank character in string */
	int      nxtmode;
	int      dosw    = 1;
	Environ  environ;


	/* initialize environ */
	initenv(&environ);

	/* verbose: report on what's happenin' */
	if(vflag) {
			fprintf(stderr,"%sbeautifying %s%s%s\n",CYAN,GREEN,infname,NRML);
				}

	while(1) {
			nxtmode= nxtc();

				switch(nxtmode) {

						case EOF:		/* end of input file processing */
									if(outbuf[0]) output();

											if(environ.wierd) {
															if(environ.wierd & CCBPREPRO) error(XTDIO_WARNING,"ccb <%s>: end-of-file in preprocessing command\n", infname);
																		if(environ.wierd & CCBSTRING) error(XTDIO_WARNING,"ccb <%s>: end-of-file inside a string<%s>\n",infname,cprt(environ.string));
																					if(environ.wierd & CCBCOMMENT) error(XTDIO_WARNING,"ccb <%s>: end-of-file inside a comment\n", infname);
																								}
													return;

														case NXTBEGIN:	/* new input line processing */
															if(outbuf[0]) output(); /* output the output buffer */

																	nb= stpblk(b);
																			if(b == buf && *nb == '#' && !(environ.wierd & (CCBSTRING|CCBCOMMENT))) {
																							char *bb;

																										bb            = stpblk(nb+1);
																													environ.wierd|= CCBPREPRO;

																																if(bb[0] == 'i' && bb[1] == 'f') {		/* #if #ifdef #ifndef	*/
																																					pushenv(&environ);
																																									}
																																			else if(bb[0] == 'e' && bb[1] == 'l') {
																																								if(bb[2] == 's' && bb[3] == 'e') {	/* #else				*/
																																														environ= accessenv();
																																																			}
																																												else if(bb[2] == 'i' && bb[3] == 'f') {	/* #elif			*/
																																																		environ= accessenv();
																																																							}
																																																}
																																						else if(bb[0] == 'e' && b[1] == 'n' && bb[2] == 'd') {	/* #endif*/
																																											popenv();
																																															}

																																									/* ok, #... commands begin in first column, get rid of initial
																																									 * 			 * tabs (if any)
																																									 * 			 			 */
																																									for(; outb > outbuf && outb[-1] == '\t'; --outb) ;
																																												}

																					else if(environ.wierd & CCBBAR) {	/* bar comment skipper */
																									char *bb=NULL;

																												switch(environ.cmmntbar) {

																																case 1:	/* skip first newline after function */
																																					if(!strcmp(b,"\n")) {
																																											++environ.cmmntbar;
																																																continue;
																																																					}
																																									environ.cmmntbar= 0;
																																													environ.wierd  &= ~CCBBAR;
																																																	break;

																																																				case 2:	/* skip comment bar (whether - or =) */
																																																					if(b[0] == '/' && b[1] == '*' && b[2] == ' ' &&
																																																											   (b[3] == '-' || b[3] == '=')) {
																																																											for(bb= b + 3; *bb; ++bb)
																																																																	  if(*bb != '-' && *bb != '=') break;
																																																																}
																																																									if(bb && bb[0] == ' ' && bb[1] == '*' && bb[2] == '/' &&
																																																															   bb[3] == '\n') {
																																																															++environ.cmmntbar;
																																																																				b= bb + 3;
																																																																									continue;
																																																																														}
																																																													environ.cmmntbar= 0;
																																																																	environ.wierd  &= ~CCBBAR;
																																																																					break;

																																																																								case 3:	/* skip final newline */
																																																																									if(!strcmp(b,"\n")) {
																																																																															environ.cmmntbar= 0;
																																																																																				environ.wierd  &= ~CCBBAR;
																																																																																									continue;
																																																																																														}
																																																																													environ.cmmntbar= 0;
																																																																																	environ.wierd  &= ~CCBBAR;
																																																																																					break;
																																																																																									}
																															}

																							/* Exceptions: case/default and private/public/protected
																							 * 		 * Detab by one.
																							 * 		 		 */
																							else if(cflag && !environ.wierd && (
																												   (nb[0] == 'c' &&
																												    			nb[1] == 'a' &&
																																		nb[2] == 's' &&
																																					nb[3] == 'e' && !(isalpha(nb[4]) || isdigit(nb[4]) || nb[4] == '_'))  ||
																												   		   (nb[0] == 'd' &&
																														    			nb[1] == 'e' &&
																																				nb[2] == 'f' &&
																																							nb[3] == 'a' &&
																																									    nb[4] == 'u' &&
																																									    			nb[5] == 'l' &&
																																															nb[6] == 't' && !(isalpha(nb[7]) || isdigit(nb[7]) || nb[7] == '_'))  ||
																														   		   (nb[0] == 'p' &&
																																    			nb[1] == 'r' &&
																																						nb[2] == 'i' &&
																																									nb[3] == 'v' &&
																																												nb[4] == 'a' &&
																																															nb[5] == 't' &&
																																																		nb[6] == 'e' && !(isalpha(nb[7]) || isdigit(nb[7]) || nb[7] == '_'))  ||
																																   		   (nb[0] == 'p' &&
																																		    			nb[1] == 'u' &&
																																								nb[2] == 'b' &&
																																											nb[3] == 'l' &&
																																														nb[4] == 'i' &&
																																																	nb[5] == 'c' && !(isalpha(nb[6]) || isdigit(nb[6]) || nb[6] == '_'))  ||
																																		   		   (nb[0] == 'p' &&
																																				    			nb[1] == 'r' &&
																																										nb[2] == 'o' &&
																																													nb[3] == 't' &&
																																																nb[4] == 'e' &&
																																																			nb[5] == 'c' &&
																																																						nb[6] == 't' &&
																																																									nb[7] == 'e' &&
																																																												nb[8] == 'd' && !(isalpha(nb[9]) || isdigit(nb[9]) || nb[9] == '_')) )){
																																																																b   = nb;
																																																																			strcpy(outbuf,environ.tabs);
																																																																						outb= outbuf + (environ.tab - environ.tabs);
																																																																									if(environ.tabs[0] == '\t') --outb;
																																																																												}

																									/* blank lines should only have a newline in them */
																									else if(*b == '\n') {
																													outbuf[0]= '\0';
																																outb     = outbuf;
																																			}

																											else {
																															strcpy(outbuf,environ.tabs);
																																		outb= outbuf + (environ.tab - environ.tabs);
																																					}

																													/* fall through */

																												case NXTNRML:	/* normal: not new input, not end of file */
																													if(environ.wierd) {

																																	/* assume that the character switch which follows the "wierd"
																																	 * 			 * block is to be not-done
																																	 * 			 			 */
																																	dosw= 0;

																																				if(environ.wierd & CCBCOMMENT) {
																																									if(b[0] == '*' && b[1] == '/') { /* end-of-comment */
																																															*(outb++)     = *(b++);
																																																				*(outb++)     = *b;
																																																									environ.wierd&= ~CCBCOMMENT;
																																																														}
																																													else *(outb++)= *b;
																																																	}
																																							else if(environ.wierd & CCBSTRING) {
																																												if(*b == '\\') {
																																																		*(outb++)= *(b++);
																																																							*(outb++)= *b;
																																																												}
																																																else if(*b == environ.string) {
																																																						*(outb++)     = *b;
																																																											environ.string= ' ';
																																																																environ.wierd&= ~CCBSTRING;
																																																																					}
																																																				else if(*b == '\n')
																																																									  error(XTDIO_ERROR,"ccb <%s>: newline in string\n\t%ld:%s\n",
																																																											  				  infname,linenum,sprt(buf));
																																																								else (*outb++)= *b;
																																																												}
																																										else if(environ.wierd & CCBPREPRO) {
																																															if(b[0] == '\\' && b[1] == '\n') {
																																																					/* stay in preprocess mode for next line */
																																																					strcpy(outb,b);
																																																										outb+= strlen(outb);
																																																															++b;
																																																																				}
																																																			else if(b[0] == '\n') {
																																																									/* leave preprocessing mode */
																																																									*(outb++)     = *b;
																																																														environ.wierd&= ~CCBPREPRO;
																																																																			}
																																																							else *(outb++)= *b;
																																																											}
																																													else if((environ.wierd & CCBIFWHILE) ||
																																																		        (environ.wierd & CCBDO)      ||
																																																						        (environ.wierd & CCBBAR)      ) dosw= 1;
																																																else error(XTDIO_ERROR,
																																																					 "ccb <%s>: environ.wierd(%d<%s%s%s%s%s%s>) is wierd\n\t%ld:%s\n",
																																																					 			  infname,
																																																								  			  environ.wierd,
																																																											  			  (environ.wierd & CCBPREPRO)?  "prepro "   : "",
																																																														  			  (environ.wierd & CCBSTRING)?  "string "   : "",
																																																																	  			  (environ.wierd & CCBCOMMENT)? "cmmnt "    : "",
																																																																				  			  (environ.wierd & CCBIFWHILE)? "ifwhile "  : "",
																																																																							  			  (environ.wierd & CCBDO)?      "do "       : "",
																																																																										  			  (environ.wierd & CCBBAR)?     "bar "      : "",
																																																																													  			  linenum,
																																																																																  			  sprt(buf));
																																																			}	/* wierd stuff */
																															else dosw= 1;


																																	/* normal character processing */
																																	if(dosw) switch(*b) {

																																				case '(':	/* )( */
																																								++environ.paren;
																																											(*outb++)   = *b;
																																														break;

																																																case ')':
																																																	--environ.paren;
																																																				(*outb++)= *b;
																																																							if(environ.paren < 0)
																																																											  error(XTDIO_ERROR,"ccb <%s>: mismatched parentheses\n\t%ld:%s\n",
																																																													  			  infname,linenum,sprt(buf));
																																																										else if(environ.func == NOTFUNC && environ.paren == 0 &&
																																																															        environ.curly == 0) environ.func= MAYBEFUNC;
																																																													else if(environ.paren == 0 && (environ.wierd & CCBIFWHILE)) {
																																																																		if(fixcurly(&environ)) indent('{',&environ); /* } */
																																																																	environ.wierd&= ~CCBIFWHILE;
																																																																					}
																																				break;

																																						case '{':	/* }{ */
																																							if(environ.curly == 0 && environ.func == MAYBEFUNC) {
																																												while(outb > outbuf && isspace(outb[-1])) --outb;
																																																if(outb > outbuf) (*outb++)   = '\n';
																																																				(*outb++)   = *b;
																																																								(*outb++)   = '\n';
																																																												b           = stpblk(b+1) - 1;
																																																																environ.func= INFUNC;
																																																																				}
																																										else (*outb++)= *b;
																																													indent('{',&environ);	/* } */
																																																break;

																																																		case '}':
																																																			indent(*b,&environ);

																																																						if(environ.curly < 0)
																																																										  error(XTDIO_ERROR,"ccb <%s>: mismatched {}\n\t%ld:%s\n",
																																																												  			  infname,linenum,sprt(buf));
																																																									else if(environ.curly == 0) {
																																																														char *ob;

																																																																		if(environ.func == INFUNC) {

																																																																								/* force newline before function-ending curly brace */
																																																																								for(ob= outb-1; ob > outbuf && isspace(*ob); --ob) ;
																																																																													if(ob > outbuf) {
																																																																																				outb     = ob + 1;
																																																																																										(*outb++)= '\n';
																																																																																																}
																																																																																		(*outb++)= *b;
																																																																																							(*outb++)= '\n';

																																																																																												/* skip trailing white space */
																																																																																												b= stpblk(b+1) - 1;
																																																																																																	if(*b != '\n') {
																																																																																																								strcpy(outb,environ.tabs);
																																																																																																														outb+= (environ.tab - environ.tabs);
																																																																																																																				}

																																																																																																						/* print out a between-function comment bar */
																																																																																																						if(bflag) {
																																																																																																													int ibar;
																																																																																																																			(*outb++)= '\n';
																																																																																																																									(*outb++)= '/';
																																																																																																																															(*outb++)= '*';
																																																																																																																																					(*outb++)= ' ';
																																																																																																																																											for(ibar= 0; ibar < bflag; ++ibar) (*outb++)= '-';
																																																																																																																																																	(*outb++)= ' ';
																																																																																																																																																							(*outb++)= '*';
																																																																																																																																																													(*outb++)= '/';
																																																																																																																																																																			(*outb++)= '\n';
																																																																																																																																																																									(*outb++)= '\n';
																																																																																																																																																																															environ.wierd  |= CCBBAR;
																																																																																																																																																																																					environ.cmmntbar= 1;
																																																																																																																																																																																											}
																																																																																																											}
																																																																						else (*outb++)= *b;
																																																																										environ.func= NOTFUNC;
																																																																														}
																																																												else if(iflag) {
																																																																	char *bb;
																																																																					(*outb++)= *b;
																																																																									bb       = stpblk(b+1);
																																																																													if(*bb) {
																																																																																			(*outb++)= '\n';
																																																																																								strcpy(outb,environ.tabs);
																																																																																													outb    += (environ.tab - environ.tabs);
																																																																																																		b        = bb - 1;
																																																																																																							}
																																																																																	}
																																																															else (*outb++)= *b;
																																																																		break;

																																																																				case '/':
																																																																					if(b[1] == '*') {		/* C style comment		*/
																																																																										environ.wierd|= CCBCOMMENT;
																																																																														(*outb++)= *(b++);
																																																																																		(*outb++)= *b;
																																																																																						}
																																																																								else if(b[1] == '/') {	/* C++ style comment	*/
																																																																													strcpy(outb,b);
																																																																																	b   += strlen(b) - 1;
																																																																																					outb+= strlen(outb);
																																																																																									}
																																																																											else (*outb++)= *b;		/* not a comment		*/
																																																																														break;

																																																																																case '"':	/* go into "..." string mode */
																																																																																case '\'':	/* go into '...' string mode */
																																																																																	environ.string= *b;
																																																																																				environ.wierd|= CCBSTRING;
																																																																																							(*outb++)     = *b;
																																																																																										break;

																																																																																												default:
																																																																																													switch(environ.func) {

																																																																																																	case MAYBEFUNC:
																																																																																																						if     (isalpha(*b)) environ.func= INFUNC;
																																																																																																										else if(!isspace(*b)) {
																																																																																																																if(cppmode) {
																																																																																																																							if(*b != ':' && *b != ',') environ.func= NOTFUNC;
																																																																																																																													}
																																																																																																																					else environ.func= NOTFUNC;
																																																																																																																										}
																																																																																																														(*outb++)= *b;
																																																																																																																		break;

																																																																																																																					case NOTFUNC:
																																																																																																																						(*outb++)= *b;
																																																																																																																										break;

																																																																																																																													case INFUNC:
																																																																																																																														if(iflag) {	/* test for if/do/while/etc. processing */

																																																																																																																																				/* "if" */
																																																																																																																																				if(b[0] == 'i' && b[1] == 'f' &&
																																																																																																																																											  (b == buf || isspace(b[-1]) || ispunct(b[-1])) &&
																																																																																																																																											  					  (isspace(b[2]) || ispunct(b[2]))) {
																																																																																																																																											environ.wierd|= CCBIFWHILE;
																																																																																																																																																	(*outb++)     = *(b++);
																																																																																																																																																							(*outb++)     = *b;
																																																																																																																																																													}

																																																																																																																																									/* "for" */
																																																																																																																																									else if(b[0] == 'f' && b[1] == 'o' && b[2] == 'r' &&
																																																																																																																																																  (b == buf || isspace(b[-1]) || ispunct(b[-1]))  &&
																																																																																																																																																  					  (isspace(b[3]) || ispunct(b[3]))) {
																																																																																																																																																environ.wierd|= CCBIFWHILE;
																																																																																																																																																						(*outb++)     = *(b++);
																																																																																																																																																												(*outb++)     = *(b++);
																																																																																																																																																																		(*outb++)     = *b;
																																																																																																																																																																								}

																																																																																																																																														/* "while" */
																																																																																																																																														else if(b[0] == 'w' && b[1] == 'h' && b[2] == 'i' &&
																																																																																																																																																					   b[3] == 'l' && b[4] == 'e' &&
																																																																																																																																																					   					   (b == buf || isspace(b[-1]) || ispunct(b[-1])) &&
																																																																																																																																																										   					   (isspace(b[5]) || ispunct(b[5]))) {
																																																																																																																																																					environ.wierd|= CCBIFWHILE;
																																																																																																																																																											(*outb++)     = *(b++);
																																																																																																																																																																	(*outb++)     = *(b++);
																																																																																																																																																																							(*outb++)     = *(b++);
																																																																																																																																																																													(*outb++)     = *(b++);
																																																																																																																																																																																			(*outb++)     = *b;
																																																																																																																																																																																									}

																																																																																																																																																			/* do */
																																																																																																																																																			else if(b[0] == 'd' && b[1] == 'o'               &&
																																																																																																																																																										  (b == buf || isspace(b[-1]) || ispunct(b[-1])) &&
																																																																																																																																																										  					  (isspace(b[2]) || ispunct(b[2]))) {
																																																																																																																																																										(*outb++)     = *(b++);
																																																																																																																																																																(*outb++)     = *b;
																																																																																																																																																																						if(fixcurly(&environ)) indent('{',&environ); /* } */
																																																																																																																																																																												}

																																																																																																																																			/* "else" */
																																																																																																																																			else if(b[0] == 'e' && b[1] == 'l'               &&
																																																																																																																																										  b[2] == 's' && b[3] == 'e'                     &&
																																																																																																																																										  					  (b == buf || isspace(b[-1]) || ispunct(b[-1])) &&
																																																																																																																																															  					  (isspace(b[4]) || ispunct(b[4]))) {
																																																																																																																																										(*outb++)     = *(b++);
																																																																																																																																																(*outb++)     = *(b++);
																																																																																																																																																						(*outb++)     = *(b++);
																																																																																																																																																												(*outb++)     = *b;
																																																																																																																																																																		if(fixcurly(&environ)) indent('{',&environ); /* } */
																																																																																																																																																																								}

																																																																																																		/* "switch" */
																																																																																																		else if(b[0] == 's' && b[1] == 'w' && b[2] == 'i' &&
																																																																																																									   b[3] == 't' && b[4] == 'c' && b[5] == 'h'      &&
																																																																																																									   					   (b == buf || isspace(b[-1]) || ispunct(b[-1])) &&
																																																																																																														   					   (isspace(b[6]) || ispunct(b[6]))) {
																																																																																																									environ.wierd|= CCBIFWHILE;
																																																																																																															(*outb++)     = *(b++);
																																																																																																																					(*outb++)     = *(b++);
																																																																																																																											(*outb++)     = *(b++);
																																																																																																																																	(*outb++)     = *(b++);
																																																																																																																																							(*outb++)     = *(b++);
																																																																																																																																													(*outb++)     = *b;
																																																																																																																																																			}

																																																																																																							/* normal -- just copy it over */
																																																																																																							else (*outb++)= *b;
																																																																																																												}
					else (*outb++)= *b;
									break;
													}
			break;
						}	/* switch on *b */
		break;
				}		/* switch on nxtmode */

	}	/* while loop */
}

/* ------------------------------------------------------------------------ */

/* indent: sets up tabs */
static void indent(char curly,Environ *environ)
{
	int old_curly;

	/* sanity */
	if(!environ->tab) environ->tab= environ->tabs;

	/* current indentation level */
	old_curly= environ->tab - environ->tabs;

	if     (curly == '{') ++environ->curly;
	else if(curly == '}') --environ->curly;


	/* fflag: de-tab inside of functions by one tab */
	if(fflag && environ->func == INFUNC) ++old_curly;

	/* adjust leading tab string */
	while(old_curly < environ->curly && environ->tab < environ->tabs + BUFSIZE-1) {
			++old_curly;
				*(environ->tab++)= '\t';
					*environ->tab    = '\0';
						}
	while(old_curly > environ->curly && environ->tab > environ->tabs) {
			--old_curly;
				*--environ->tab= '\0';
					}

}

/* ------------------------------------------------------------------------ */
static Environ *ehd = NULL;
static Environ *etl = NULL;

/* pushenv: push an Environ on stack */
static void pushenv(Environ *environ)
{
	Environ        *envnxt;
	Environ        *envprv;


	double_link(Environ,ehd,etl,"Environ");


	/* copy environ into new stacked Environ, but preserve nxt/prv pointers */
	envnxt  = etl->nxt;
	envprv  = etl->prv;
	*etl    = *environ;
	etl->nxt= envnxt;
	etl->prv= envprv;

}

/* ------------------------------------------------------------------------ */

/* popenv: pop an Environ */
static void popenv()
{

	delete_double_link(Environ,etl,ehd,etl);


}

/* ------------------------------------------------------------------------ */

/* accessenv: access Environ on tail of stack */
static Environ accessenv()
{
	Environ env;


	if(!etl) {
			env.nxt    = env.prv= NULL;
				env.string = 0;
					env.tab    = NULL;
						env.tabs[0]= '\0';
							env.curly  = -1;
								env.paren  = -1;
									env.wierd  = 0;
										error(XTDIO_WARNING,"accessenv <%s>: attempt to access an empty Environ stack\n",infname);
											return env;
												}

	env    = *etl;
	env.nxt= env.prv= NULL;

	return env;
}

/* --------------------------------------------------------------------- */

/* output: handles output */
static void output()
{
	char *ob;

	if(wflag) return;


	if(outb > outbuf) {
			/* strip off trailing blanks and tabs */
			for(--outb; *outb == ' ' || *outb == '\t'; --outb) ;
				(*++outb)= '\0';						/* terminate outbuf			*/

					/* final output stage */
					if(dflag) {	/* { */
								/* de-tab tab-tab...tab} by one tab */
								ob= stpblk(outbuf); /* { */
										if(*ob == '}' && ob > outbuf) fputs(outbuf+1,fpout);
												else                          fputs(outbuf,fpout);
														}
						else fputs(outbuf,fpout);
							}
					outb= outbuf;

	}

	/* ------------------------------------------------------------------------ */

	/* initenv: this function initializes an Environ */
	static void initenv(Environ *environ)
	{

		environ->string  = ' ';
		environ->tabs[0] = '\0';
		environ->tab     = environ->tabs;
		environ->cmmntbar= 0;
		environ->curly   = 0;
		environ->func    = NOTFUNC;
		environ->paren   = 0;
		environ->wierd   = CCBNRML;
		environ->nxt     = environ->prv= NULL;

	}

	/* ------------------------------------------------------------------------ */


